#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import http
import logging
import pathlib
import sys
from distutils.util import strtobool
from typing import Tuple, Union

import requests
from ingest_envoy.interfaces import IngestLauncherIF
from ingest_envoy.launchers import (
    IngestCalibrationLauncher,
    IngestImageLauncher,
    IngestObservationLauncher,
)
from ingest_envoy.solicitor import Solicitor
from ingest_envoy.utilities import IngestType, VLASSIngestType
from pycapo import CapoConfig

"""
Setup and Launch ingestion via Workspaces
"""

logger = logging.getLogger("ingest_envoy")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def _get_settings(
    arg_type: Union[IngestType, VLASSIngestType], filename: str = None, cal_spl: str = None, source_dir: str = None
) -> dict:
    """
    Determine needed settings from metadata.json file

    :param arg_type: the type of ingestion to run
    :param filename: the provided metadata file
    :param cal_spl: the calibration science product locator to use (needed for VLASS image ingestion)
    :param source_dir: the product source directory to ingest (needed for VLASS ingestion)
    :return: a dict containing required settings to run the provided ingestion type
    """
    ingestion_settings = CapoConfig().settings("edu.nrao.workspaces.IngestionSettings")
    script_location = CapoConfig().settings("edu.nrao.workspaces.ProcessingSettings").scriptLocation
    capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").externalServiceUrl
    workflow_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

    parameters = Solicitor(arg_type, [capability_url, workflow_url], filename=filename).solicit_parameters()

    if cal_spl is not None:
        parameters["calSpl"] = cal_spl
    if source_dir is not None:
        parameters["workflowDir"] = pathlib.Path(source_dir).name
        parameters["sourcePath"] = pathlib.Path(source_dir).parent

    parameters["staging_area"] = ingestion_settings.stagingDirectory
    parameters["storage_area"] = ingestion_settings.storageDirectory
    parameters["useIngest"] = strtobool(ingestion_settings.useIngest)
    parameters["script_location"] = script_location
    parameters["workflowUrl"] = workflow_url

    return parameters


def arg_parser() -> argparse.ArgumentParser:
    """
    Create option parser for this envoy

    :return: argument parser
    """
    parser = argparse.ArgumentParser(
        description="Workspaces Ingestion System", formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "--calibration",
        nargs=1,
        action="store",
        required=False,
        help="run ingestion for a calibration product",
    )
    parser.add_argument(
        "--image",
        nargs=1,
        action="store",
        required=False,
        help="run ingestion for an image product",
    )
    parser.add_argument(
        "--observation",
        nargs=2,
        action="store",
        required=False,
        help="run ingestion for an observation",
    )
    parser.add_argument(
        "--seci",
        nargs=2,
        action="store",
        required=False,
        help="run ingestion for VLASS SECI image products",
    )
    return parser


def check_ingest_type(args_type: Union[IngestType, VLASSIngestType], parameters: dict) -> bool:
    """
    Ensure that we are running the correct type of ingestion for a given product

    :param args_type: type of ingestion requested
    :param parameters: the parameters determined from metadata.json
    :return: boolean representing if the requested ingestion type matches the product type
    """
    if "workflowName" not in parameters and args_type == IngestType.OBS:
        return True

    wf_name = parameters["workflowName"]
    if args_type.value in wf_name:
        return True
    elif args_type == IngestType.IMG and "imaging" in wf_name:
        return True
    return False


def set_cal_status_flag(parameters: dict):
    """
    Send AMQP message to AAT to set calibration status on EB.

    :param parameters: dictionary of available input parameters
    :return:
    """
    request_id = parameters["requestId"]

    url = parameters["workflowUrl"] + f"/workflows/requests/message-archive/cal-ingestion-complete"

    response = requests.post(url, json={"request_id": request_id, "project_code": parameters["project"]})

    if response.status_code != http.HTTPStatus.OK:
        logger.warning(
            f"WARNING: Failed to set calibration status flag in archive for request #{request_id}."
            f" Please set the flag manually!"
        )


def send_aat_image_message(parameters: dict, ingest_type: Union[IngestType, VLASSIngestType]):
    """
    Send AMQP message to AAT to set off image indexing

    :param parameters: dictionary of available input parameters
    :param ingest_type: type of ingestion that was run
    :return:
    """
    url = parameters["workflowUrl"] + f"/workflows/requests/message-archive/img-ingestion-complete"
    if isinstance(ingest_type, VLASSIngestType):
        payload = {"project_code": parameters["project"]}
    else:
        payload = {"request_id": parameters["requestId"], "project_code": parameters["project"]}

    response = requests.post(url, json=payload)

    if response.status_code != http.HTTPStatus.OK:
        logger.warning(
            f"WARNING: Failed to send image ingestion complete to archive for image associated with "
            f"calibration {parameters['calSpl']} and project {parameters['project']}."
            f" Please set off the index manually!"
        )


def send_aat_obs_message(parameters: dict):
    """
    Send AMQP message to AAT to set off observation (VLBA) indexing

    :param parameters: dictionary of available input parameters
    :return:
    """
    url = parameters["workflowUrl"] + f"/workflows/requests/message-archive/obs-ingestion-complete"
    payload = {"project_code": parameters["project"]}

    response = requests.post(url, json=payload)

    if response.status_code != http.HTTPStatus.OK:
        logger.warning(
            f"WARNING: Failed to send observation ingestion complete to archive for observation associated with "
            f"project {parameters['project']}."
            f" Please set off the index manually!"
        )


def _determine_ingestion_info(
    args: argparse.Namespace,
) -> Tuple[Union[IngestType, VLASSIngestType], dict, IngestLauncherIF]:
    """
    Determine parameters and launcher needed for ingestion

    :param args: args dict from parser
    :return: tuple containing the ingestion type, parameters dict, and launcher object
    """
    arg_type = None
    parameters = None
    launcher = None

    if args.calibration is not None:
        arg_type = IngestType.CAL
        parameters = _get_settings(arg_type, filename=args.calibration[0])
        launcher = IngestCalibrationLauncher(parameters)
    elif args.image is not None:
        arg_type = IngestType.IMG
        parameters = _get_settings(arg_type, filename=args.image[0])
        launcher = IngestImageLauncher(arg_type, parameters)
    elif args.seci is not None:
        arg_type = VLASSIngestType.SECI
        parameters = _get_settings(arg_type, cal_spl=args.seci[0], source_dir=args.seci[1])
        launcher = IngestImageLauncher(arg_type, parameters)
    if args.observation is not None:
        arg_type = IngestType.OBS
        parameters = _get_settings(arg_type, filename=args.observation[0], source_dir=args.observation[1])
        launcher = IngestObservationLauncher(parameters)

    return arg_type, parameters, launcher


def main():
    """
    Entry point for envoy

    :return:
    """
    args = arg_parser().parse_args()

    arg_type, parameters, launcher = _determine_ingestion_info(args)

    if check_ingest_type(args_type=arg_type, parameters=parameters):
        result = launcher.launch_ingestion()
        if result == 0:
            # Ingestion succeeded
            logger.info("Ingestion finished successfully! Sending status message to archive...")
            if arg_type == IngestType.CAL:
                set_cal_status_flag(parameters)
            elif arg_type == IngestType.OBS:
                send_aat_obs_message(parameters)
            else:
                send_aat_image_message(parameters, arg_type)
        else:
            logger.error(f"ERROR: {arg_type} ingestion failure! Please check logs")
            sys.exit(1)
    else:
        logger.error(
            f"ERROR: The workflow request to be ingested does not match specified "
            f"ingestion type {arg_type}. Exiting...."
        )
        sys.exit(1)
